//------------------------------------------------------------------------------
//
//  stencil shadow : gyab
//
//------------------------------------------------------------------------------
#include <stdio.h>
#include <windows.h>
#include <ctype.h>
#include <math.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <vector>

#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")
#pragma comment(lib, "dinput8.lib")
#pragma warning(disable:4244)
#pragma warning(disable:4305)

#define D3DCLEAR_DEFAULT   (D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL)
#define RELEASE(x)         { if(x) (x)->Release(); x = NULL; }
#define ScreenX            800
#define ScreenY            600

void InitScene();
void TermScene();
void RenderScene();

HWND                  hWnd;
WNDCLASSEX            wcex;
LPDIRECT3D9           d3d9;
LPDIRECT3DDEVICE9     d3ddevice;
static D3DPRESENT_PARAMETERS d3dpp = {
  ScreenX, ScreenY, D3DFMT_X8R8G8B8, 1,   D3DMULTISAMPLE_NONE, 0,
  D3DSWAPEFFECT_DISCARD, 0, !FALSE, TRUE, D3DFMT_D24S8, 0, 0, TRUE,
};

static LRESULT WINAPI WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
  int temp = wParam & 0xFFF0;
  switch(msg) {
  case WM_SYSCOMMAND:
    if(temp == SC_MONITORPOWER || temp == SC_SCREENSAVE) {
      return 0;
    }
    break;
  case WM_IME_SETCONTEXT:
    lParam &= ~ISC_SHOWUIALL;
    break;
  case WM_CLOSE:
  case WM_DESTROY:
    PostQuitMessage(0);
    return 0;
    break;
  case WM_KEYDOWN:
    if(wParam == VK_ESCAPE) PostQuitMessage(0);
    break;
  }
  return DefWindowProc(hWnd, msg, wParam, lParam);
}


int main(int a, char *b[]) {
  const LPSTR    AppName = "tekitou stencil shadow";
  MSG  msg;
  static WNDCLASSEX wcex = {
    sizeof(WNDCLASSEX), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW,
    WindowProc, 0L, 0L, GetModuleHandle(NULL), LoadIcon(NULL, IDI_APPLICATION),
    LoadCursor(NULL, IDC_ARROW), (HBRUSH)GetStockObject(BLACK_BRUSH),
    NULL, AppName, NULL
  };

  if(!RegisterClassEx(&wcex)) return 1;
  DWORD style   = d3dpp.Windowed  ? WS_OVERLAPPEDWINDOW : WS_POPUP;
  DWORD exstyle = WS_EX_APPWINDOW | (d3dpp.Windowed ? WS_EX_WINDOWEDGE : 0);
  POINT point   = {(GetSystemMetrics( SM_CXSCREEN ) - ScreenX ) / 2, (GetSystemMetrics( SM_CYSCREEN ) - ScreenY) / 2};
  RECT  rc      = {0, 0, ScreenX, ScreenY};
  AdjustWindowRectEx( &rc, style, FALSE, exstyle);
  hWnd = CreateWindowEx(exstyle, AppName, AppName, style,
    point.x, point.y, rc.right - rc.left, rc.bottom - rc.top, NULL, NULL, wcex.hInstance, NULL);
  if(!hWnd) {
    OutputDebugString("ERROR : Can't Create Window");
    return 0;
  }

  ShowWindow(hWnd, SW_SHOW);
  UpdateWindow(hWnd);
  d3dpp.Windowed ? (void)0 : ShowCursor(FALSE);

  d3d9 = Direct3DCreate9(D3D_SDK_VERSION);
  if(!d3d9) {
    OutputDebugString("ERROR : Direct3DCreate9");
    return 0;
  }
  d3d9->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &d3ddevice);
  if(d3ddevice){
    BOOL bDone = FALSE;
    InitScene();
    do {
      while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
        if(msg.message == WM_QUIT) {
          bDone = TRUE;
          break;
        } else {
          TranslateMessage(&msg);
          DispatchMessage(&msg);
        }
      }
      do {
        HRESULT hr = E_FAIL;
        hr = d3ddevice->TestCooperativeLevel();
        if(hr == D3DERR_DEVICELOST) {
          break;
        }
        if(hr == D3DERR_DEVICENOTRESET) {
          d3ddevice->Reset(&d3dpp);
        }
        if(d3ddevice->BeginScene() == D3D_OK) {
          RenderScene();
          d3ddevice->EndScene();
        }
        d3ddevice->Present( NULL, NULL, NULL, NULL );
      } while(0);
      Sleep(0);
    } while(!bDone);
  }
  TermScene();
  RELEASE(d3ddevice);
  RELEASE(d3d9);
  return 0;
}


//---------------//---------------//---------------//---------------//---------------//---------------
//Shadow Mesh
//---------------//---------------//---------------//---------------//---------------//---------------
#define  MESH_MAX  16
LPD3DXMESH vmesh[MESH_MAX] = {0};

struct MESH_VTX {
  D3DXVECTOR3 p;
  D3DXVECTOR3 n;
};

struct MESH_INFO {
  DWORD       Id;
  D3DXMATRIX  mWorld;
};

std::vector<MESH_INFO> vMeshInfo;

void AddMesh(DWORD Id, D3DXMATRIX &m) {
  MESH_INFO info;
  info.Id = Id;
  info.mWorld = m;
  vMeshInfo.push_back(info);
}

void AddCube(D3DXMATRIX &m) {
  AddMesh(0, m);
}

struct SHADOW_VTX {
  D3DXVECTOR3 p[3];
};

std::vector<SHADOW_VTX> sbatch;
void ResetSBatch() {
  sbatch.clear();
}

void AddSPoly(D3DXVECTOR3 v[3], D3DXVECTOR3 &Light) {
  const FLOAT scale = 256.0;
  D3DXVECTOR3 sv[3] = { v[0] + Light * scale, v[1] + Light * scale, v[2] + Light * scale, };
  SHADOW_VTX svtx;
  svtx.p[0] = v[0]; svtx.p[1] =  v[1]; svtx.p[2] = sv[0]; sbatch.push_back(svtx);
  svtx.p[0] = v[1]; svtx.p[1] = sv[1]; svtx.p[2] = sv[0]; sbatch.push_back(svtx);

  svtx.p[0] = v[0]; svtx.p[1] = sv[0]; svtx.p[2] =  v[2]; sbatch.push_back(svtx);
  svtx.p[0] = v[2]; svtx.p[1] = sv[0]; svtx.p[2] = sv[2]; sbatch.push_back(svtx);

  svtx.p[0] = v[1]; svtx.p[1] =  v[2]; svtx.p[2] = sv[2]; sbatch.push_back(svtx);
  svtx.p[0] = v[1]; svtx.p[1] = sv[2]; svtx.p[2] = sv[1]; sbatch.push_back(svtx);
}

void BuildSBatch(LPD3DXMESH mesh, D3DXVECTOR3 &Light) {
  if(mesh) {
    MESH_VTX *pv    = NULL;
    WORD     *idx   = NULL;
    DWORD     count = mesh->GetNumFaces();
    DWORD     Flags = D3DLOCK_READONLY | D3DLOCK_NOOVERWRITE;
    mesh->LockVertexBuffer(Flags, (LPVOID *)&pv);
    mesh->LockIndexBuffer(Flags, (LPVOID *)&idx);
    for(int i = 0 ; i < count; i++) {
      D3DXVECTOR3 c;
      D3DXVECTOR3 vt0 = pv[idx[3*i+0]].p;
      D3DXVECTOR3 vt1 = pv[idx[3*i+1]].p;
      D3DXVECTOR3 vt2 = pv[idx[3*i+2]].p;
      D3DXVec3Cross(&c, &(vt2 - vt1), &(vt1 - vt0));
      D3DXVec3Normalize(&c, &c);
      if(D3DXVec3Dot(&c, &Light) > -0.01) { //eps
        D3DXVECTOR3 v[3] = { vt0, vt1, vt2, };
        AddSPoly(v, Light);
      }
    }
    mesh->UnlockIndexBuffer();
    mesh->UnlockVertexBuffer();
  }
}

void DrawSBatch(D3DXVECTOR3 &Light) {
  std::vector<MESH_INFO>::iterator ite = vMeshInfo.begin();
  std::vector<MESH_INFO>::iterator end = vMeshInfo.end();
  FLOAT Lx = Light.x;
  FLOAT Ly = Light.y;
  FLOAT Lz = Light.z;
  while(ite != end) {
    LPD3DXMESH  mesh = vmesh[ite->Id];
    if(mesh) {
      D3DXMATRIX  m = ite->mWorld;
      D3DXVECTOR3 L;
      D3DXMatrixInverse(&m, NULL, &ite->mWorld );
      //D3DXVec3TransformCoord(&L, &Light, &m);
      L.x = Lx*m._11 + Ly*m._21 + Lz*m._31;// + m._41;
      L.y = Lx*m._12 + Ly*m._22 + Lz*m._32;// + m._42;
      L.z = Lx*m._13 + Ly*m._23 + Lz*m._33;// + m._43;

      ResetSBatch();
      BuildSBatch(mesh, L);
      if(sbatch.size() > 0)
      {
        d3ddevice->SetTransform(D3DTS_WORLD, &ite->mWorld);
        d3ddevice->SetFVF(D3DFVF_XYZ);
        d3ddevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, sbatch.size(), &(sbatch[0]), sizeof(SHADOW_VTX) / 3);
      }
      ResetSBatch();
    }
    ite++;
  }
}


void ResetBatch() {
  vMeshInfo.clear();
}

void DrawBatch() {
  std::vector<MESH_INFO>::iterator ite = vMeshInfo.begin();
  std::vector<MESH_INFO>::iterator end = vMeshInfo.end();
  while(ite != end) {
    d3ddevice->SetTransform(D3DTS_WORLD, &ite->mWorld);
    if(vmesh[ite->Id]) vmesh[ite->Id]->DrawSubset(0);
    ite++;
  }
}

void DrawCube(int Id,  FLOAT tx, FLOAT ty, FLOAT tz, FLOAT sx, FLOAT sy, FLOAT sz, FLOAT rx, FLOAT ry, FLOAT rz) {
  D3DXMATRIX mWorld, mScale, mRot;
  D3DXMatrixTranslation(&mWorld, tx, ty, tz);
  D3DXMatrixRotationYawPitchRoll(&mRot, ry, rx, rz);
  D3DXMatrixScaling(&mScale, sx, sy, sz);
  mWorld = mScale * mRot * mWorld;
  AddMesh(Id, mWorld);
}

void DrawScene() {
  d3ddevice->SetRenderState(D3DRS_ZENABLE,             TRUE);
  d3ddevice->SetRenderState(D3DRS_ZWRITEENABLE,        TRUE);
  d3ddevice->SetRenderState(D3DRS_CULLMODE,            D3DCULL_CCW);
  DrawBatch();
}

void DrawShadow(D3DCOLOR color, D3DXVECTOR3 &Light) {
  struct SHADOW_RECT {
    FLOAT v[4];
    D3DCOLOR color;
  };
  SHADOW_RECT vs[] = {
    0,       0,       0, 1, color,
    ScreenX, 0,       0, 1, color,
    ScreenX, ScreenY, 0, 1, color,
    0,       ScreenY, 0, 1, color,
  };
  
  //Stencil Shadow
  d3ddevice->SetRenderState(D3DRS_STENCILENABLE,       TRUE);
  d3ddevice->SetRenderState(D3DRS_ZENABLE,             TRUE);
  d3ddevice->SetRenderState(D3DRS_ZWRITEENABLE,        FALSE);
  d3ddevice->SetRenderState(D3DRS_COLORWRITEENABLE,    0);
  d3ddevice->SetRenderState(D3DRS_TWOSIDEDSTENCILMODE, TRUE);//dx9
  d3ddevice->SetRenderState(D3DRS_CULLMODE,            D3DCULL_NONE);
  d3ddevice->SetRenderState(D3DRS_STENCILFUNC,         D3DCMP_ALWAYS);
  d3ddevice->SetRenderState(D3DRS_STENCILPASS,         D3DSTENCILOP_INCR);
  d3ddevice->SetRenderState(D3DRS_STENCILZFAIL,        D3DSTENCILOP_KEEP);
  d3ddevice->SetRenderState(D3DRS_STENCILFAIL,         D3DSTENCILOP_KEEP);
  d3ddevice->SetRenderState(D3DRS_CCW_STENCILFUNC,     D3DCMP_ALWAYS);
  d3ddevice->SetRenderState(D3DRS_CCW_STENCILPASS,     D3DSTENCILOP_DECR);
  d3ddevice->SetRenderState(D3DRS_CCW_STENCILZFAIL,    D3DSTENCILOP_KEEP);
  d3ddevice->SetRenderState(D3DRS_CCW_STENCILFAIL,     D3DSTENCILOP_KEEP);
  d3ddevice->SetRenderState(D3DRS_SHADEMODE,           D3DSHADE_FLAT);
  DrawSBatch(Light);
  d3ddevice->SetRenderState(D3DRS_SHADEMODE,           D3DSHADE_GOURAUD);
  d3ddevice->SetRenderState(D3DRS_TWOSIDEDSTENCILMODE, FALSE );
  d3ddevice->SetRenderState(D3DRS_CULLMODE,            D3DCULL_CCW);
  d3ddevice->SetRenderState(D3DRS_ZWRITEENABLE,        TRUE);
  
  //Draw Shadow Rect
  d3ddevice->SetRenderState(D3DRS_ZENABLE,             FALSE );
  d3ddevice->SetRenderState(D3DRS_CULLMODE,            D3DCULL_CCW);
  d3ddevice->SetRenderState(D3DRS_COLORWRITEENABLE,    0xF);
  d3ddevice->SetRenderState(D3DRS_STENCILREF,          0x1);
  d3ddevice->SetRenderState(D3DRS_STENCILFUNC,         D3DCMP_LESSEQUAL );
  d3ddevice->SetRenderState(D3DRS_STENCILPASS,         D3DSTENCILOP_KEEP);

  //Draw Rect
  d3ddevice->SetRenderState(D3DRS_ALPHABLENDENABLE,    TRUE);
  d3ddevice->SetRenderState(D3DRS_SRCBLEND,            D3DBLEND_SRCALPHA );
  d3ddevice->SetRenderState(D3DRS_DESTBLEND,           D3DBLEND_INVSRCALPHA );
  d3ddevice->SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE);
  d3ddevice->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vs, sizeof(SHADOW_RECT));
  d3ddevice->SetRenderState(D3DRS_STENCILENABLE,       FALSE);
  d3ddevice->SetRenderState(D3DRS_ALPHABLENDENABLE,    FALSE);
}

//---------------//---------------//---------------//---------------//---------------//---------------

static FLOAT       angle = 0;
static FLOAT       px    = 0;
static FLOAT       py    = 0;
static FLOAT       pz    = 0;

void SetupScene() {
  D3DXMATRIX mView, mProj;
  D3DXMatrixLookAtLH(&mView, &D3DXVECTOR3(px,  py, pz), &D3DXVECTOR3(0,  1.5,  0), &D3DXVECTOR3(0,  1,  0));
  D3DXMatrixPerspectiveFovLH(&mProj, D3DX_PI / 3.0f, (FLOAT)ScreenX / (FLOAT)ScreenY, 1.0f / 32.0f, 256.0f);
  d3ddevice->SetTransform( D3DTS_VIEW, &mView );
  d3ddevice->SetTransform( D3DTS_PROJECTION, &mProj );
  d3ddevice->SetRenderState( D3DRS_LIGHTING, FALSE ); //disable specure ambient
  
  //gnd
  DrawCube(0,  0, 0, 0, 10, 1, 10, 0, 0, 0);
  
  DrawCube(0,  -0,   1.7, -0,   0.5, 0.5, 0.5, 0.3 + angle * 3, 0 + angle * 3, 0 + angle * 3);
  DrawCube(0,  -2,   1.5, -2,   1.0, 1.0, 1.0, 1.0,             0.0,           0.0);
  DrawCube(0,  -2.5, 1.7,  2.5, 1.0, 1.0, 1.0, 0.0+angle,       1.0,           0.0);
  DrawCube(0,  2,    1.4, -2,   1.0, 1.0, 1.0, 0.0,             0.0+angle,     1.0);
  DrawCube(0,  2,      1,  2,   1.0, 1.0, 1.0, 0.0,             0.0,           0.0+angle);
  
  FLOAT kk = 0.3-angle;
  FLOAT rt = 0+angle*2;
  FLOAT sc = 0.35;
  int   div = 32;
  for(int i = 0;  i < div; i++) {
    sc = 0.25 + cos(FLOAT(i)*5) * 0.15;
    DrawCube(1, sin(kk)*2.5, 2.0+sin(kk*3),cos(kk)*4, sc, sc, sc, 0.3+rt, 0.4+rt, 0.5+rt);
    kk += 3.14*2/FLOAT(div);
    rt += 0.3;
  }
}

void TermScene() {
  ResetBatch();
  for(int i = 0 ; i < MESH_MAX; i++) {
    RELEASE(vmesh[i]);
  }
}

void InitScene() {
  TermScene();
  D3DXCreateBox(d3ddevice, 2, 2, 2, &vmesh[0], NULL);
  D3DXCreateSphere(d3ddevice, 1, 4, 4, &vmesh[1], NULL);
}

void RenderScene() {
  angle += 0.005;
  py = 5;
  px = sin(angle) * 3;
  pz = cos(angle) * 5;

  d3ddevice->Clear( 0, NULL, D3DCLEAR_DEFAULT, 0xFFFFE8C8, 1.0f, 0);
  ResetBatch();
  SetupScene();
  
  //Draw Normal Scene
  DrawScene();
  
  //Draw Shadow Scene
  int   ite   = 16; //OMOI
  FLOAT an    = 0;
  for(int i = 0 ; i < ite; i++) {
    d3ddevice->Clear( 0, NULL, D3DCLEAR_STENCIL, 0xFFFFFFFF, 1.0f, 0);
    D3DXVECTOR3 Light = D3DXVECTOR3(-sin(an), -1.7+cos(an * 47)*0.7, cos(an));
    D3DXVec3Normalize(&Light, &Light);
    DrawShadow(0x08F07040, Light);
    an += 3 / FLOAT(ite);
  }
}

